# -*- coding: utf-8 -*-
"""
Collection Library
"""


# Imports
# =====================================================================


import os
from pathlib import PosixPath

# from pprint import pprint

import anyconfig
from json_schema_for_humans.generation_configuration import GenerationConfiguration
from json_schema_for_humans.template_renderer import TemplateRenderer

from json_schema_for_humans.generate import copy_additional_files_to_target
from json_schema_for_humans.schema.schema_to_render import SchemaToRender

from cafram.utils import (
    to_yaml,
    to_json,
    write_file,
    read_file,
    first,
)
import paasify.errors as error

from paasify.framework import PaasifyObj, FileLookup
from paasify.stack_components import StackAssembler

# from paasify.common import get_paasify_pkg_dir


# Global variables
# =====================================================================


DOC_APP_1 = """
# {name}

Documentation for app: `{name}`

## Documentation

{app_readme}


## Variables

``` yaml title="vars.yml"
{app_vars}
```

## Docker compose files



"""

DOC_APP_2 = """

### {name}


``` yaml title="{source}"
{dockerfile}
```


"""


DOC_JSONNET_1 = """
# {requested_ident}

{name}


{description}

## Documentation

{meta_readme}


## Metadata

``` yaml
{meta_yaml}
```

## Tag documentation

This documentation is autogenerated from the plugin source code. The single
html page is [here]({slug}.html).

<script type="text/javascript" src="https://code.jquery.com/jquery-1.8.3.js"></script>
<script type="text/javascript" src="/paasify/static/paasify.js"></script>

<iframe id="schemadoc" src="../{slug}.html"
width="100%" height="600px"
frameborder="0"
overflow="hidden"
/>

"""


DOC_SITE_1 = """
# {title}


{app_readme}


## Applications

{app_list_fragment}

## Jsonnet plugins

{jsonnet_list_fragment}


---

<sup>Made with :heart: by [mrjk](https://github.com/mrjk) for [Paasify](https://github.com/barbu-it/paasify)</sup>

"""


# Helpers
# =====================================================================


def gen_schema_doc(schema_file, dest_dir, filename, config=None, **kwargs):
    "Generate documentation for a given schema"

    # Prepare config
    final_config = dict(
        minify=True,
        deprecated_from_description=False,
        default_from_description=False,
        expand_buttons=False,
        link_to_reused_ref=False,
        # copy_css=copy_css,
        # copy_js=copy_js,
        # config=config_file,
        # config_parameters=config,
    )
    user_config = config or {}
    final_config.update(user_config)
    config = GenerationConfiguration(**final_config)

    dest_html = os.path.join(dest_dir, filename)
    schema_to_render = SchemaToRender(
        PosixPath(schema_file),  # Source schema
        PosixPath(dest_html),  # Destination file
        PosixPath(dest_dir),  # Destination dir
    )

    template_renderer = TemplateRenderer(config)
    schema_to_render.render(template_renderer, None)
    copy_additional_files_to_target([schema_to_render], config)

    return dest_html


# Main Application class
# =====================================================================


class CollectionScanner(PaasifyObj):
    """A Collection Scanner Object

    Useful for scanning the content of a collection

    requirement: one path to lookup
    """

    # ident = "CollectionScanner"

    def __init__(self, *args, parent=None, path=None, **kwargs):

        # print ("RECEIVED", args, kwargs)

        # super().__init__(*args, **kwargs)
        self.log = parent.log

        self.path_raw = path
        self.jsonnets = None
        self.apps = None

        # Init object
        self.scan_collection(path)

    def scan_collection(self, path):
        "Find available apps for a given collection path"

        search_for_jsonnet = False
        match_apps = {}
        match_jsonnets = []

        # Check for apps
        for root, dirs, files in os.walk(path):
            for name in dirs:
                if name.startswith("_"):
                    pass
                elif name == ".paasify":
                    search_for_jsonnet = True
                else:
                    target = os.path.join(path, name)
                    match_apps[name] = self.scan_app(target)

            # Itereate one leve deep only, so we quit here
            # Maybe we should find for all docker-compose.yml ...
            break

        # Check for jsonnet plugins
        if search_for_jsonnet:
            plugin_dir = os.path.join(path, ".paasify", "plugins")
            match_jsonnets = self.scan_jsonnets(plugin_dir)

        # Return results
        self.apps = match_apps
        self.jsonnets = match_jsonnets
        return {
            "jsonnets": match_jsonnets,
            "apps": match_apps,
        }

    def scan_app_file(self, path, name):
        "Scan a single app file to determine what it is"

        ret = None
        file_dir = os.path.join(path, name)
        file_ = name
        if file_.startswith("_"):
            # We ignore everything starting by _
            pass
        elif file_ in ["README.md"]:
            ret = {
                "type": "doc",
                "dir": file_dir,
                "name": "main",
            }
        elif file_.endswith("md"):
            name = file_.replace(".md", "")
            ret = {
                "type": "doc",
                "dir": file_dir,
                "name": name,
            }
        elif file_ in ["vars.yml"]:
            ret = {
                "type": "vars",
                "dir": file_dir,
                "name": "main",
            }
        elif file_.startswith("docker-compose"):
            if file_ == "docker-compose.yml" or file_ == "docker-compose.yaml":
                ret = {
                    "type": "docker_file",
                    "dir": file_dir,
                    "name": "main",
                }
            else:
                name = file_.replace(".yml", "")
                name = name.replace("docker-compose.", "")
                ret = {
                    "type": "docker_file",
                    "dir": file_dir,
                    "name": name,
                }
        elif file_.endswith("jsonnet"):
            name = file_.replace(".jsonnet", "")
            ret = {
                "type": "jsonnet_app",
                "dir": file_dir,
                "name": name,
            }

        return ret

    def scan_app(self, path, name=None):
        "Scan a single app"

        if not os.path.isdir(path):
            return None

        name = name or os.path.basename(path)
        matches = []
        for root, dirs, files in os.walk(path):
            for file_ in files:
                ret = self.scan_app_file(path, file_)
                if ret:
                    matches.append(ret)

            break
        return matches

    def scan_jsonnets(self, path):
        "Scan single dir for jsonnets"

        if not os.path.isdir(path):
            return None

        matches = []
        for root, dirs, files in os.walk(path):

            for file_ in files:
                ret = None
                file_dir = os.path.join(path, file_)
                if file_.startswith("_"):
                    pass
                elif file_.endswith("jsonnet"):
                    name = file_.replace(".jsonnet", "")
                    ret = {
                        "type": "jsonnet",
                        "dir": file_dir,
                        "name": name,
                    }
                if ret:
                    matches.append(ret)
            break
        return matches


class CollectionDocumator(PaasifyObj):
    """A Collection Documator Object

    Generate documentation for a path that represent a collection

    requirement: one path to lookup,
    optional : dest_dir,  otherwhise inside .paasify/docs/
    """

    ident = "DocGen"

    DEFAULT_DOC_DIR = "_docs"

    def __init__(self, parent=None, path=None, dest_dir=None, name=None, **kwargs):

        self.log = parent.log
        # super().__init__(*args, **kwargs)

        # print ("NEW COLLECTION", path)
        path_abs = os.path.abspath(path or ".")

        self.path = path
        self.dest_dir = dest_dir
        self.name = str(name or os.path.basename(path_abs)).strip(os.sep)

        # print ("CURRENT ANEM", self.name, path)
        self.scanner = CollectionScanner(parent=self, ident=self.name, path=path)

    def generate(
        self, format=None, path=None, dest_dir=None, singleton=True, mkdocs_config=None
    ):
        "Generate collection documentation"

        path = path or self.path
        dest_dir = dest_dir or self.dest_dir

        assert path, f"Path does not exists: {path}"
        if not os.path.exists(path):
            assert False, f"Path does not exists: {path}"

        # Prepare stack builder
        sta = StackAssembler(parent=self, ident=f"{self}")
        dest_dir = dest_dir or os.path.join(path, ".paasify", self.DEFAULT_DOC_DIR)

        # Ensure that dest dir exists
        if not os.path.isdir(dest_dir):
            self.log.notice(f"Creating parent directories: {dest_dir}")
            os.makedirs(dest_dir)

        jsonnet_list = self._generate_jsonnet(sta, dest_dir)
        app_list = self._generate_apps(dest_dir)
        self._generate_index_page(
            dest_dir, app_list=app_list, jsonnet_list=jsonnet_list
        )
        if mkdocs_config:
            self._generate_mkdoc_config(mkdocs_config)

        self.log.notice(f"Documentation generated in: {dest_dir}")

    def retitle(self, md_text, lvl):
        "Increase title levels"
        char = "#"
        ret = []
        for line in md_text.split("\n"):

            if line.startswith(char):
                line = char * lvl + line
            ret.append(line)
        return "\n".join(ret)

    def _generate_apps(self, dest_dir):
        "Generate apps doc"

        apps = self.scanner.apps

        # Loop over each jsonnet
        ret = []
        for app_name, files in apps.items():

            self.log.info(f"Generate app doc '{app_name}' having {len(files)} file(s)")

            slug = f"app_{app_name}"
            main_components = [file_ for file_ in files if file_.get("name") == "main"]

            # README and vars
            file_readme = first([f for f in main_components if f.get("type") == "doc"])
            file_vars = first([f for f in main_components if f.get("type") == "vars"])

            app_conf = {}
            app_readme = "No readme"
            app_vars = "No var file"
            if file_readme:
                app_readme = read_file(file_readme["dir"])
                app_readme = self.retitle(app_readme, 2)
            if file_vars:
                app_vars = read_file(file_vars["dir"])
                conf = anyconfig.load(file_vars["dir"])
                app_conf = conf or {}

            # Generate docker compose doc fragements
            docker_files = [f for f in files if f.get("type") == "docker_file"]

            if len(docker_files) < 1:
                self.log.warning(f"Skip invalid app: {app_name}")
                continue

            # docker_file_main = [ret for ret in docker_files if ret.get("name") == "main" ]
            index = next(
                (
                    i
                    for i, item in enumerate(docker_files)
                    if item.get("name") == "main"
                ),
                -1,
            )
            docker_file_main = [docker_files.pop(index)]

            extra = []
            loop = docker_file_main + docker_files
            for docker_file in loop:

                content = read_file(docker_file["dir"])
                source = os.path.basename(docker_file["dir"])

                frag = DOC_APP_2.format(
                    name=docker_file["name"], source=source, dockerfile=content
                )

                extra.append(frag)

            # Generate top fragment
            _dest = f"{dest_dir}/{slug}.md"
            vars_ = {
                "name": app_name,
                "requested_ident": app_name,
                "kind": "app",
                "app_vars": app_vars,
                "app_readme": app_readme,
                "md_file": _dest,
            }
            vars_.update(app_conf)
            app_doc = DOC_APP_1.format(**vars_)
            app_doc = app_doc + "\n".join(extra)

            # Write output

            self.log.info(f"Write app markdown doc: {_dest}")
            write_file(_dest, app_doc)
            ret.append(vars_)

        return ret

    def _generate_jsonnet(self, sta, dest_dir):
        "Generate jsonnect schema doc"

        jsonnets = self.scanner.jsonnets

        # Prepare json-schemadoc
        config = dict(
            minify=False,
            deprecated_from_description=True,
            default_from_description=False,
            expand_buttons=False,
            link_to_reused_ref=False,
            # copy_css=copy_css,
            # copy_js=copy_js,
            # config=config_file,
            # config_parameters=config,
        )

        # Loop over each jsonnet
        ret = []
        for jsonnet in jsonnets:
            self.log.info(f"Generating doc for {jsonnet}")

            # Prepare vars
            # =========================
            name = jsonnet["name"]
            type_ = jsonnet["type"]
            jsonnet_desc = jsonnet.get("description")
            slug = f"{type_}_{name}"
            jsonnet_file = jsonnet["dir"]
            # pkg_dir = get_paasify_pkg_dir()
            # conf_file = os.path.join(pkg_dir, "assets", f"jsondoc_jsonnet_{type_}.yml")

            # Parse jsonnet metadata
            # =========================
            params = {
                "args": {},
            }
            try:
                plugin_meta = sta.process_jsonnet_exec(
                    jsonnet_file,
                    "metadata",
                    params,
                    # import_dirs=jsonnet_lookup_dirs,
                )
            except error.JsonnetBuildFailed:
                self.log.error(f"Skip invalid jsonnet: {jsonnet}")
                continue

            # Extract plugin metadata
            json_schema = plugin_meta.pop("jsonschema", {})
            md_readme = plugin_meta.pop(
                "readme", "No documentation provided by the plugin"
            )
            md_readme = self.retitle(md_readme, 2)
            yaml_meta = to_yaml(plugin_meta)

            # Create markdown doc
            # =========================
            vars_ = {
                "requested_ident": name,
                "name": name,
                "description": jsonnet_desc,
                "slug": slug,
                "meta_readme": md_readme,
                "meta_yaml": yaml_meta,
                "kind": type_,
                # name=name,
                # slug=slug,
                # meta_readme=md_readme,
                # meta_yaml=to_yaml(plugin_meta),
                # kind=type_,
            }
            vars_.update(plugin_meta)

            # Generated various docs
            md_doc = DOC_JSONNET_1.format(**vars_)
            _dest = f"{dest_dir}/{slug}.md"
            self.log.info(f"Write markdown doc: {_dest}")
            write_file(_dest, md_doc)

            _dest_schema = f"{dest_dir}/{slug}.json"
            self.log.info(f"Write json schema: {_dest_schema}")
            write_file(_dest_schema, to_json(json_schema))

            # Update results
            vars_.update(
                {
                    "md_file": _dest,
                    "json_schema_file": _dest_schema,
                }
            )
            ret.append(vars_)

            # Generate jsonnet schema doc
            # =========================
            # dest_html = os.path.join(dest_dir, f"{slug}.{ext}")
            gen_schema_doc(_dest_schema, dest_dir, f"{slug}.html", config=config)

        return ret

    def _generate_index_page(self, dest_dir, app_list=None, jsonnet_list=None):
        "Generate a singleton module documentation"

        app_list = app_list or []
        jsonnet_list = jsonnet_list or []
        path = self.path
        pattern = ["README.md", "README.txt"]

        # Generate README
        app_readme = "No readme"
        file_readme = FileLookup(path=path, pattern=pattern).paths(first=True)
        if file_readme:
            app_readme = read_file(file_readme)
            app_readme = self.retitle(app_readme, 2)

        # Generate app fragments
        app_list_fragment = ["List of applications: ", ""]
        for app in app_list:
            fname = os.path.basename(app["md_file"])
            # app_ident = app["requested_ident"]
            app_name = app["name"]
            app_product = app.get("app_product", None)
            app_desc = app.get("app_description", app.get("homepage_description", ""))
            app_image = app.get("app_image", "")
            app_image_name = app.get("app_image_name")
            app_image_version = app.get("app_image_version")

            # app_icon = app.get("app_icon")

            if app_image_name:
                app_image = app_image_name
                if app_image_version:
                    app_image = f"{app_image}:{app_image_version}"

            app_sup = ""
            if app_image:
                app_sup += app_image
            if len(app_sup) > 0:
                app_sup = f" <sup>({app_sup})</sup>"

            app_line = []
            if app_product and app_product != app_name:
                app_line.append(app_product)
            if app_desc:
                app_line.append(app_desc)

            app_line = (": " + ", ".join(app_line)) if len(app_line) > 0 else ""
            if len(app_sup) > 0:
                app_line += app_sup

            line = f"  * [`{app_name}`]({fname}){app_line}"
            app_list_fragment.append(line)

        # Generate jsonnet fragments
        jsonnet_list_fragment = ["List of jsonnet plugins:", ""]
        for jsonnet in jsonnet_list:
            fname = os.path.basename(jsonnet["md_file"])
            jsonnet_ident = jsonnet["requested_ident"]
            jsonnet_name = jsonnet["name"]
            jsonnet_desc = jsonnet.get("description", None)

            jsonnet_line = (jsonnet_desc or jsonnet_name) if jsonnet_desc else ""
            jsonnet_line = (": " + jsonnet_line) if len(jsonnet_line) > 0 else ""
            line = f"  * [`{jsonnet_ident}`]({fname}){jsonnet_line}"
            jsonnet_list_fragment.append(line)

        # Generate format vars
        vars_ = {
            "name": self.name,
            "title": "Overview",
            "app_readme": app_readme,
            "app_list_fragment": "\n".join(app_list_fragment),
            "jsonnet_list_fragment": "\n".join(jsonnet_list_fragment),
        }

        # Generate doc
        col_doc = DOC_SITE_1.format(**vars_)
        _dest = f"{dest_dir}/index.md"
        self.log.info(f"Write collection markdown doc: {_dest}")
        write_file(_dest, col_doc)

    def _generate_mkdoc_config(self, dest_dir):
        "Generate mkdocs config"

        self.log.info("Gen mkdocs not implemented yet")



# -*- coding: utf-8 -*-
"""
Paasify errors
"""


class PaasifyError(Exception):
    """Base class for other exceptions"""

    paasify = True
    rc = 1

    def __init__(self, message, rc=None, advice=None):
        # self.paasify = True
        self.advice = advice

        # pylint: disable=invalid-name
        self.rc = rc or self.rc
        super().__init__(message)


class ProjectNotFound(PaasifyError):
    """Raised when project is not found"""

    rc = 17


class ProjectInvalidConfig(PaasifyError):
    """Raised when project config contains errors"""

    rc = 18


class ShellCommandFailed(PaasifyError):
    """Raised when project config contains errors"""

    rc = 18


class StackNotFound(PaasifyError):
    """Raised when stack is not found"""

    rc = 19


class StackMissingOrigin(PaasifyError):
    """Raised when a stack origin is not determined"""

    rc = 20


class DockerBuildConfig(PaasifyError):
    "Raised when docker-config failed"
    rc = 30


class DockerCommandFailed(PaasifyError):
    "Raised when docker-config failed"
    rc = 32


class JsonnetBuildFailed(PaasifyError):
    "Raised when jsonnet failed"
    rc = 31


class DockerUnsupportedVersion(PaasifyError):
    "Raised when docker-config failed"
    rc = 33


class JsonnetProcessError(PaasifyError):
    "Raised when jsonnet file can't be executed"
    rc = 34


class InvalidConfig(PaasifyError):
    "Raised when invalid syntax for config"
    rc = 36


class PaasifyNestedProject(PaasifyError):
    "Raised when a project is created into an existing project"
    rc = 35


class StackMissingDockerComposeFile(PaasifyError):
    """Raised when a stack can't find a docker-compose.yml"""

    rc = 38


class BuildStackFirstError(PaasifyError):
    """Raised when a trying to interact with stack but docker-compose.yml is missing"""

    rc = 39


class OnlyOneStackAllowed(PaasifyError):
    """Raised when trying to apply command one more than one stack"""

    rc = 41


class YAMLError(PaasifyError):
    """Raised when having issues with YAML file"""

    rc = 42


class MissingTag(PaasifyError):
    """Raised when referencing unexistant tag"""

    rc = 43


class MissingApp(PaasifyError):
    """Raised when referencing unknown app"""

    rc = 44


class ConfigBackendError(PaasifyError):
    """Raised when could not work with cafram"""

    rc = 45


class InvalidSourceConfig(PaasifyError):
    """Raised when a source is not configured properly"""

    rc = 46


class UndeclaredVariable(PaasifyError):
    """Raised when a a variable is called but not defined"""

    rc = 47


class MissingFile(PaasifyError):
    """Raised when referencing unexisting file"""

    rc = 48


class DuplicateSourceName(PaasifyError):
    """Raised when two sources have the same name"""

    rc = 49



# -*- coding: utf-8 -*-
"""Paasify Source management libray

"""

import os
import logging
from pprint import pprint, pformat  # noqa: F401

from giturlparse import parse as gitparse

from cafram.utils import _exec, first
from cafram.nodes import NodeMap, NodeList

from paasify.common import get_paasify_pkg_dir
from paasify.framework import PaasifyObj, FileReference
import paasify.errors as error

log = logging.getLogger(__name__)


# =====================================================================
# Source management
# =====================================================================


class Source(NodeMap, PaasifyObj):
    """A Source instance"""

    conf_default = {
        "remote": None,
        "name": None,
        "prefix": None,
        "dir": None,
        "install": None,
        # "prefix": "https://github.com/%s.git",
    }

    schema_def = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "title": "Paasify Source configuration",
        "oneOf": [
            {
                "type": "string",
            },
            {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "name": {
                        "type": "string",
                    },
                    "remote": {
                        "type": "string",
                    },
                    "ref": {
                        "type": "string",
                    },
                    "prefix": {
                        "type": "string",
                    },
                    "aliases": {
                        "type": "array",
                    },
                },
            },
        ],
    }

    def extract_short(self, short):
        "Guess data from source short forms"
        ret = None

        # Try if remote
        remote = self.remote or short
        if remote:
            out = gitparse(remote)
            if hasattr(out, "url"):
                # Guess from valid URL
                ret = {
                    "remote": out.url,
                    "name": out.repo,
                    "dir": f"{out.owner}-{out.repo}",
                    "install": "clone",
                    "guess": f"a git repo: {out.url}",
                }
            else:
                self.log.trace(f"Source is not a git repo: {self._node_conf_raw}")

        # Try if dir
        value_ = self.dir or self.remote or self.name or short
        if not ret and value_:
            # So is it a path? Relative, absolute?
            install_ = self.install or "none"

            # Look for direct path
            path = FileReference(value_, root=self.runtime.root_path, keep=True)
            if os.path.isdir(path.path()):
                path_ = path.path()
                ret = {
                    "remote": None,
                    "name": str(self.name or os.path.basename(path.path_abs())),
                    "path": path_,
                    "install": install_,
                    "guess": f"path: {path_}",
                }
            else:
                self.log.trace(
                    f"Source is not found in path '{path}' for {self._node_conf_raw}"
                )

            # Look into <prj>.paasify/collections
            if not ret:
                path_ = FileReference(
                    value_, self.runtime.project_collection_dir, keep=True
                ).path()
                if os.path.isdir(path_):
                    ret = {
                        "remote": None,
                        "name": str(self.name or os.path.basename(path_)),
                        "path": path_,
                        "install": install_,
                        "guess": f"inside the stack: {path_}",
                    }
                else:
                    self.log.trace(
                        f"Source is not found in collection dir '{path}' for {self._node_conf_raw}"
                    )

        # Try default gh repo
        value_ = self.dir or self.remote or short
        if not ret and value_:
            # Last change to guess from pattern
            owner = None
            repo = value_
            if "/" in value_:
                parts = short.split("/", 2)
                owner = parts[0]
                repo = parts[1]

            if not owner:
                owner = "paasify"

            remote_ = f"https://github.com/{owner}/{repo}.git"
            ret = {
                "remote": remote_,
                "name": f"{owner}/{repo}",
                "dir": f"{owner}/{repo}",
                "install": "clone",
                "guess": "Refer to github repository ({remote_})",
            }

        if not ret:
            msg = f"Invalid configuration for source: {self._node_conf_raw}"
            raise error.InvalidSourceConfig(msg)

        assert ret["install"]
        return ret

    def node_hook_init(self, **kwargs):
        "Setup object vars from parents"
        self.obj_prj = self._node_parent._node_parent
        self.collection_dir = self.obj_prj.runtime.project_collection_dir

    def node_hook_transform(self, payload):
        "Transform short form into dict"

        if isinstance(payload, str):
            payload = {"short": payload}
        return payload

    def node_hook_final(self):
        "Init source correctly from available elements"

        self.runtime = self._node_parent._node_parent.runtime
        self.relative = self.runtime.relative

        # TODO: this is very awful
        self.short = getattr(self, "short", None)
        self.name = getattr(self, "name", None)
        self.dir = getattr(self, "dir", None)
        self.aliases = getattr(self, "aliases", [])
        self.remote = getattr(self, "remote", None)
        self.install = getattr(self, "install", None)
        self.path = None

        extracted = None
        extracted = self.extract_short(self.short)
        for key, value in extracted.items():
            setattr(self, key, getattr(self, key, None) or value)
        self.log.debug(f"Source '{self.name}' is {extracted['guess']}")

        # The actual thing we all want !
        if not self.path:
            self.path = FileReference(
                self.dir, self.runtime.project_collection_dir, keep=True
            ).path()

        assert self.name
        assert self.path

    def is_git(self):
        "Return true if git repo"
        test_path = os.path.join(self.path, ".git")
        return os.path.isdir(test_path)

    def is_installed(self):
        "Return true if installed"
        return os.path.isdir(self.path)

    def cmd_install(self):
        "Install from remote"

        # Check if install dir is already present
        remote = self.remote
        install = self.install

        # Skip non installable sources
        if install == "none":
            self.log.notice(f"Source '{self.name}' is a path, no need to install")
            return

        # Ensure parent directory exists before install
        dest = os.path.join(self.collection_dir, self.dir)
        parent_dir = os.path.dirname(dest)
        if not os.path.isdir(parent_dir):
            self.log.info(f"Creating parent directories: {parent_dir}")
            os.makedirs(parent_dir)

        # Install source
        if install.startswith("symlink"):
            self.log.notice(f"Installing '{self.name}' via {install}: {remote}")

            # Ensure nothing exists first
            if os.path.exists(dest):

                if not os.path.exists(os.readlink(dest)):
                    self.log.notice(f"Removing broken link in: {dest}")
                    os.unlink(dest)
                else:
                    assert False, "Found a bug !"

            # Create the symlink
            self.log.notice(f"Collection symlinked in: {dest}")
            os.symlink(remote, dest)

        elif install == "clone":
            # Git clone that stuff
            if os.path.isdir(self.path):
                self.log.notice(
                    f"Source '{self.name}' is already installed in {self.path}"
                )
                return
            self.log.notice(f"Installing '{self.name}' git: {remote} in {self.path}")
            cli_args = ["clone", remote, self.path]
            _exec("git", cli_args, _fg=True)
        else:
            raise Exception(f"Unsupported methods: {install}")

    def cmd_update(self):
        "Update from remote"

        # Check if install dir is already present
        if not os.path.isdir(self.path):
            self.log.info("This source is not installed yet")
            return

        # Git clone that stuff
        git_url = self.git_url()
        self.log.info(f"Updating git repo: {git_url}")
        cli_args = [
            "-C",
            self.path,
            "pull",
        ]
        _exec("git", cli_args, _fg=True)


class SourcesManager(NodeList, PaasifyObj):
    "Source manager"

    conf_schema = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "title": "Paasify Source configuration",
        "description": "Configure a list of collections to install",
        "oneOf": [
            {
                "type": "null",
                "title": "Unset sources",
                "description": "If null or empty, it does not use any source",
            },
            {
                "title": "List of sources",
                "description": "Each source define a git repository and a name for stack references",
                "type": "array",
                "items": Source.schema_def,
            },
        ],
    }

    conf_children = Source

    def node_hook_transform(self, payload):

        module_path = os.path.dirname(error.__file__)
        module_path2 = get_paasify_pkg_dir()
        assert module_path == module_path2, "TODO: Validate this"
        paasify_collection = os.path.join(
            module_path, "assets", "collections", "paasify"
        )

        # Always inject the paasify core source at first
        payload = payload or []
        payload.insert(
            0,
            {
                "name": "paasify",
                "dir": paasify_collection,
            },
        )

        return payload

    def node_hook_final(self):
        "Prepare sources"

        # Look for duplicate names
        known_names = []
        for src in self.get_children():

            if src.name in known_names:
                dups = [dup._node_conf_raw for dup in self.find_by_name_alias(src.name)]
                dups = ", ".join(dups)
                msg = f"Two stacks have the same name: {dups}"
                raise error.DuplicateSourceName(msg)
            known_names.append(src.name)

            for alias in src.aliases:
                if alias in known_names:
                    dups = [
                        dup._node_conf_raw for dup in self.find_by_name_alias(alias)
                    ]
                    dups = ", ".join(dups)
                    msg = f"An alias have duplicate assignments: {dups}"
                    raise error.DuplicateSourceName(msg)
                known_names.append(alias)

        # Assign a default node:
        default = self.find_by_name_alias("default")
        if len(default) < 1:
            first_node = first(self.get_children())
            if first_node:
                first_node.aliases.append("default")
                self.log.info(
                    f"First source has been set as default: {first_node.name}"
                )
            else:
                self.log.info("No source found for this project")

    def get_all(self):
        "Return the list of all sources"
        return list(self.get_children())

    def list_all_names(self) -> list:
        "Return a list of valid string names"

        sources = self.get_children()
        return [src.name for src in sources]

    def get_app_source(self, app_name, source=None):
        "Return the source of an app_name"

        if source:
            sources = self.find_by_name_alias(source)
        else:
            sources = self.get_all()

        checked_dirs = []
        for src in sources:
            check_dir = os.path.join(src.path, app_name)
            if os.path.isdir(check_dir):
                return src
            checked_dirs.append(check_dir)

        # Fail and explain why
        sources = [src.name for src in self.get_all()]
        src_str = ",".join(sources)
        msg_dirs = " ".join(checked_dirs)
        msg = f"Impossible to find app '{app_name}' in sources '{src_str}' in paths: {msg_dirs}"

        if len(sources) != checked_dirs:
            hint = "Are you sure you have installed sources before? If not, run: paasify src install"
            self.log.warning(hint)
        raise error.MissingApp(msg)

    def get_search_paths(self):
        "Get search paths"

        return [{"path": src.path, "src": src} for src in self.get_children()]

    def find_by_name_alias(self, string, name=None, alias=None):
        "Return all src that match name or alias"

        name = name or string
        alias = alias or string

        sources = self.get_children()
        return [src for src in sources if name == src.name or alias in src.aliases]

    # Commands
    # ======================

    def cmd_tree(self) -> list:
        "Show a tree of all sources"

        print("Not implemented yet")

    def cmd_ls(self, explain=False) -> list:
        "List all stacks names"

        sources = self.get_children()
        for src in [None] + sources:

            if not src:
                remote = "REMOTE"
                name = "NAME"
                is_installed = "INSTALLED"
                is_git = "GIT"
                path = "PATH"
            else:
                is_installed = "True" if src.is_installed() else "False"
                is_git = "True" if src.is_git() else "False"
                remote = src.remote or ""
                name = src.name
                path = src.path

            print(
                f"  {name :<20} {is_installed :<10} {is_git :<8} {remote :<50} {path}"
            )

    def cmd_install(self):
        "Ensure all sources are installed"

        sources = self.get_children()
        for src in sources:
            log.notice(f"Installing source: {src.name}")
            src.cmd_install()

    def cmd_update(self):
        "Update all source"

        sources = self.get_children()
        for src in sources:
            log.notice(f"Installing source: {src.name}")
            src.cmd_update()



# -*- coding: utf-8 -*-
"""Paasify common library

Holds common pieces of code

"""

import os
import logging
from enum import Enum
from string import Template

# from pprint import pprint

# =====================================================================
# Init
# =====================================================================
log = logging.getLogger()


class OutputFormat(str, Enum):
    "Available paasify format outputs"

    # pylint: disable=invalid-name

    yaml = "yaml"
    json = "json"
    # toml = "toml"


class SchemaTarget(str, Enum):
    "Available schema items"

    app = "app"
    prj = "prj"
    prj_config = "prj_config"
    prj_stacks = "prj_stacks"
    prj_sources = "prj_sources"


# =====================================================================
# Misc functions
# =====================================================================


def uniq(seq):
    """Remove duplicate duplicates items in a list while preserving order"""
    return list(dict.fromkeys(seq))


def update_dict(dict1, dict2, strict=False):
    """Update dict1 keys with null value from dict2"""
    result = dict1.copy()
    for key, new_value in dict2.items():
        value = result.get(key)
        if not strict and not value:
            result[key] = new_value
        elif strict and value is None:
            result[key] = new_value

    return result


def list_parent_dirs(path):
    """
    Return a list of the parents paths
    path treated as strings, must be absolute path
    """
    result = [path]
    val = path
    while val and val != os.sep:
        val = os.path.split(val)[0]
        result.append(val)
    return result


def find_file_up(names, paths):
    """
    Find every files names in names list in
    every listed paths
    """
    assert isinstance(names, list), f"Names must be array, not: {type(names)}"
    assert isinstance(paths, list), f"Paths must be array, not: {type(names)}"

    result = []
    for path in paths:
        for name in names:
            file_path = os.path.join(path, name)
            if os.access(file_path, os.R_OK):
                result.append(file_path)

    return result


def filter_existing_files(root_path, candidates):
    """Return only existing files"""
    result = [
        os.path.join(root_path, cand)
        for cand in candidates
        if os.path.isfile(os.path.join(root_path, cand))
    ]
    return list(set(result))


def cast_docker_compose(var):
    "Convert any types to strings"

    if var is None:
        result = ""
    elif isinstance(var, (bool)):
        result = "true" if var else "false"
    elif isinstance(var, (str, int)):
        result = str(var)
    elif isinstance(var, list):
        result = ",".join(var)
    elif isinstance(var, dict):
        result = ",".join([f"{key}={str(val)}" for key, val in var.items()])
    else:
        raise Exception(f"Impossible to cast value: {var}")

    return result


def to_bool(string):
    "Return a boolean"
    if isinstance(string, bool):
        return string
    return string.lower() in ["true", "1", "t", "y", "yes"]


def merge_env_vars(obj):
    "Transform all keys of a dict starting by _ to their equivalent wihtout _"

    override_keys = [key.lstrip("_") for key in obj.keys() if key.startswith("_")]
    for key in override_keys:
        old_key = "_" + key
        obj[key] = obj[old_key]
        obj.pop(old_key)

    return obj, override_keys


def get_paasify_pkg_dir():
    """Return the dir where the actual paasify source code lives"""

    # pylint: disable=import-outside-toplevel
    import paasify as _

    return os.path.dirname(_.__file__)


def ensure_dir_exists(path):
    """Ensure directories exist for a given path"""
    if not os.path.isdir(path):
        log.info(f"Create directory: {path}")
        os.makedirs(path)
        return True
    return False


def ensure_parent_dir_exists(path):
    """Ensure parent directories exist for a given path"""
    parent = os.path.dirname(os.path.normpath(path))
    return ensure_dir_exists(parent)


# =====================================================================
# Class overrides
# =====================================================================


class StringTemplate(Template):
    """
    String Template class override to support version of python below 3.11

    Source code: Source: https://github.com/python/cpython/commit/dce642f24418c58e67fa31a686575c980c31dd37
    """

    def get_identifiers(self):
        """Returns a list of the valid identifiers in the template, in the order
        they first appear, ignoring any invalid identifiers."""

        ids = []
        for mo in self.pattern.finditer(self.template):
            named = mo.group("named") or mo.group("braced")
            if named is not None and named not in ids:
                # add a named group only the first time it appears
                ids.append(named)
            elif (
                named is None
                and mo.group("invalid") is None
                and mo.group("escaped") is None
            ):
                # If all the groups are None, there must be
                # another group we're not expecting
                raise ValueError("Unrecognized named group in pattern", self.pattern)
        return ids

    def is_valid(self):
        """Returns false if the template has invalid placeholders that will cause
        :meth:`substitute` to raise :exc:`ValueError`.
        """

        for mo in self.pattern.finditer(self.template):
            if mo.group("invalid") is not None:
                return False
            if (
                mo.group("named") is None
                and mo.group("braced") is None
                and mo.group("escaped") is None
            ):
                # If all the groups are None, there must be
                # another group we're not expecting
                raise ValueError("Unrecognized named group in pattern", self.pattern)
        return True


# We override this method only if version of python is below 3.11
if hasattr(Template, "get_identifiers"):
    StringTemplate = Template  # noqa: F811



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'cafram'}

packages = \
['cafram', 'paasify']

package_data = \
{'': ['*'],
 'paasify': ['assets/*',
             'assets/collections/paasify/*',
             'assets/collections/paasify/debug/*',
             'assets/collections/paasify/dns/*',
             'assets/collections/paasify/dummy/*',
             'assets/collections/paasify/example/*',
             'assets/collections/paasify/home/*',
             'assets/collections/paasify/home/icons/*',
             'assets/collections/paasify/network/*',
             'assets/collections/paasify/proxy/*']}

install_requires = \
['anyconfig>=0.13.0,<0.14.0',
 'giturlparse>=0.10.0,<0.11.0',
 'json-schema-for-humans>=0.44,<0.45',
 'jsonnet>=0.19.1,<0.20.0',
 'jsonschema>=4.17.0,<5.0.0',
 'pyaml>=21.10.1,<22.0.0',
 'rich>=13.3.1,<14.0.0',
 'ruamel-yaml>=0.17.21,<0.18.0',
 'semver>=2.13.0,<3.0.0',
 'sh>=1.14.3,<2.0.0',
 'single-version>=1.5.1,<2.0.0',
 'typer>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['paasify = paasify.cli:app']}

setup_kwargs = {
    'name': 'paasify',
    'version': '0.1.1',
    'description': 'Paasify your docker-compose files',
    'long_description': '<p align=\'center\'>\n<img src="logo/paasify_brand.svg" alt="Paasify">\n</p>\n\n<p align=\'center\'>\n<a href="https://gitter.im/barbu-it/paasify">\n<img src="https://img.shields.io/gitter/room/barbu-it/paasify" alt="Gitter"></a>\n<a href="https://pypi.org/project/paasify/">\n<img src="https://img.shields.io/pypi/v/paasify" alt="PyPI"></a>\n<a href="https://pypistats.org/packages/paasify">\n<img src="https://img.shields.io/pypi/dm/paasify" alt="PyPI - Downloads"></a>\n<a href="https://github.com/barbu-it/paasify/releases">\n<img src="https://img.shields.io/piwheels/v/paasify?include_prereleases" alt="piwheels (including prereleases)"></a>\n<a href="https://github.com/barbu-it/paasify/graphs/code-frequency">\n<img src="https://img.shields.io/github/commit-activity/m/barbu-it/paasify" alt="GitHub commit activity"></a>\n<a href="https://www.gnu.org/licenses/gpl-3.0">\n<img src="https://img.shields.io/badge/License-GPL%20v3-blue.svg" alt="License: GPL v3"></a>\n</p>\n\n<p align="center">\n<img src="https://img.shields.io/pypi/pyversions/paasify" alt="PyPI - Python Version">\n<img src="https://img.shields.io/pypi/format/paasify" alt="PyPI - Format">\n<img src="https://img.shields.io/pypi/status/paasify" alt="PyPI - Status">\n</p>\n\n-------\n\n<p align=\'center\'>\nPlease :star: this project if like it of if you want to support it!\n</p>\n\n<p align=\'center\'>\n:warning: This project is currently in alpha stage, use at your own risks! :warning:\n</p>\n\n<p align=\'center\'>\nOfficial documentation is available on [https://barbu-it.github.io/paasify/]\n</p>\n\n-------\n\n\nDeploy your docker-compose applications with ease and manage your infrastructure as code!\n\nPaasify is a Python tool that will help you to deploy large collections of `docker-compose.yml` files. It\'s an thin overlay to the `docker compose` command\nand it will generate the `docker-compose.yml` you need. It provides some ways to fetch Apps collections, to deploy them and then ensure their state\ncan be committed into version control.\n\n\nFrom an high level perspective, this looks like:\n\n<p align=\'center\'>\n<img src="docs/src/static/overview.svg" alt="Overview">\n</p>\n\n\nThis project try to overstep the missing gap between the docker-compose deployment and static code in way to achieve infrastructure as a code. If you are asking yourself on why you would use Paasify:\n\n* Manage a lot of differents `docker-compose.yml`\n* Make your `docker-compose.yml` based infrastructure [DRY](https://en.wikipedia.org/wiki/Don%27t_repeat_yourself)\n* Write large collections of `docker-compose.yml` apps once, deploy them many times\n* Integrate your apps into other services, like you can automagically add Traefik labels to your containers\n* Deploy apps in a sequential way\n* Commit your infrastructure configuration into git\n\n\n## :memo: Table of Content\n\n\n  * [Quickstart](#quickstart)\n    + [Installation with pip](#installation-with-pip)\n    + [Installation with docker](#installation-with-docker)\n    + [Usage/Examples](#usage-examples)\n    + [Demo](#demo)\n  * [Overview](#overview)\n    + [Features](#features)\n    + [Documentation](#documentation)\n    + [Requirements](#requirements)\n    + [Environment Variables](#environment-variables)\n  * [Getting help](#getting-help)\n    + [Known issues](#known-issues)\n    + [FAQ](#faq)\n    + [Support](#support)\n    + [Feedback](#feedback)\n  * [Develop](#develop)\n  * [Project information](#project-information)\n\n\n## :fire: Quickstart\n\nThere are different ways to install Paasify:\n\n* [Installation with pip](#installation-with-pip): This is the recommended installation method for people who wants to try and/or develop infrastructure.\n* [Installation with docker](#installation-with-docker): Docker installation is more recommended for production environment.\n* [Installation with git](https://barbu-it.github.io/paasify/develop/install/): If you want to improve or contribute to Paasify itself.\n\n\n### Installation with pip\n\nInstall Paasify with pip:\n\n```bash\npip install paasify\npaasify --help\n```\n\n### Installation with docker\n\nThe idea is to create a simple alias that will call the paasify image:\n\n```bash\nalias paasify=\'docker run --rm -w /work -v $PWD:/work -ti ghcr.io/barbu-it/paasify:latest paasify \'\n\npaasify --help\n```\n\n\n### Usage/Examples\n\nYou need to have docker daemon running. See requirements section for further details. Let\'s try to deploy a Hello World project with paasify. It consists in deploying a proxy, for managing incoming traffic (we uses Traefik here) and a simple page that will list all the services you deployed with paasify.\n\n```bash\npaasify new devbox\ncat <<EOF\n# My new project\n\nsources:\n  - name: community\n    remote: https://github.com/barbu-it/paasify-collection-community\n\nconfig:\n  vars:\n    app_expose_ip: 127.0.0.1\n    app_domain: ${app_expose_ip}.nip.io\n\n  tag_prefix:\n    - homepage\n    - traefik-svc\n\nstack:\n  - app: traefik\n  - app: homepage\n  - app: community:wordpress\n    tags:\n      - phpmyadmin\nEOF\n\npaasify apply\n```\n\nThen you can visit: [http://homepage.127.0.0.1.nip.io](http://homepage.127.0.0.1.nip.io). Of course you can manage your own domains and manage SSL with let\'s encrypt. To have an idea of what app you can install, please checkout official collections:\n\n* [barbu-it/paasify-collection-community](https://github.com/barbu-it/paasify-collection-community): Apps provided for and by the community\n* [barbu-it/paasify-collection-infra](https://github.com/barbu-it/paasify-collection-infra): Dev et devops oriented Apps\n\n\n## :sparkles: Overview\n\n### Features\n\n- Only use the classical syntax of docker compose\n- Allow to use any app without effort\n- Transform your own applications into collection, and publish them as git repositories\n- Allow to centralized collections into git repositories\n- Provides a powerful `docker-compose.<TAG>.yml` assemblage\n- Provides a simple but powerful variable management and templating model\n- Provides jsonnet support for more complex transformations\n- Allow to track your infrastructure changes into git\n\nPlease check the documentation to know more and see the Road Map below to see what\'s coming.\n\n\n### Documentation\n\nThe main documentation website is at [https://barbu-it.github.io/paasify/](https://barbu-it.github.io/paasify/).\n\n### Requirements\n\nThe following system requirements are:\n\n* Linux x86 based OS (not tested yet on other platforms than Linux so far)\n* `docker`\n* `docker compose` or `docker-compose`\n* `jq`\n\nFor development:\n\n* [git](https://git-scm.com/)\n* [poetry](https://python-poetry.org/)\n* [task](https://taskfile.dev/)\n\n### Environment Variables\n\nYou may use the following environment variables to adjust paasify behavior:\n\n`PAASIFY_DEBUG=false`: Show extra log levels if set to `true`\n\n`PAASIFY_TRACE=false`: Show python traces if set to `true`\n\n## :question: Getting help\n\n### Known issues\n\n* Paasify is still at this alpha stage, and not recommended (yet) for production.\n* Paasify has only been tested on Linux, more platform *may* come later.\n* Paasify heavily use the usage of docker labels, so deploying in an existing infrastructure may lead to conflicts.\n\n\n\n### FAQ\n\n#### Does paasify involve any long running services ?\n\nNope, Paasify build your `docker-compose.yml` files and do a `docker compose up`. It\'s a simple CLI program that will super-charge your `docker compose` commands.\n\n#### Is there a web UI for deployments ?\n\nNope, the intended audience of this tool is people who want to do code as infrastructure. It may be the purpose of another project tho.\n\n#### Is it possible to have it in Go?\n\nGo is a pretty good language for this kind of tool, however the author does not known Go, so it\'s too late now. Use the docker image to get a no install setup.\n\n\n### Support\n\nThere is no support outside of community support at this stage of the project. The project is still considered as immature, getting into the project as the date of today may still require you to be comfortable with programming.\n\n\n### Feedback\n\nIf you have any feedback, please open an issue.\n\n\n## :pray: Develop\n\nHere are the basic step to hack into paasify code. A more complete guide is available in the documentation.\n\n### Installation with git\n\nClone the project\n\n```bash\n  git clone https://github.com/barbu-it/paasify\n```\n\nGo to the project directory\n\n```bash\n  cd paasify\n```\n\nInstall dependencies\n\n```bash\n  task install\n```\n\n\n### Running Tests\n\nTo run tests, run the following command\n\n```bash\n  task run_tests\n```\n\nRun the quality suite\n\n```bash\n  task run_qa\n```\n\n### Contributing\n\nContributions are always welcome!\n\nSee `contributing.md` for ways to get started.\n\nPlease adhere to this project\'s `code of conduct`.\n\n\n## :earth_africa: Project information\n\n### Roadmap\n\n- Volume and secret management\n- Docker Swarm support\n\n\n### License\n\n[GNU General Public License v3.0](LICENSE.txt)\n\n### Authors\n\nThis project is brought to you thanks to Barbu-IT.\n\n- [@mrjk](https://www.github.com/mrjk)\n\n\n### Used By\n\nThis project is used by the following companies:\n\n- Barbu-IT\n\n\n### Related\n\nHere are some related projects:\n\n* [Dokku](https://github.com/dokku/dokku)\n\n\n### Support this project\n\nYou can :star: this project, contribute or donate to original author [@mrjk](https://www.github.com/mrjk):\n\n* Bitcoin: `bc1qxdtn24vl9n8e04992dwcq3pdumes0l2dqardvh`\n',
    'author': 'MrJK',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/barbu-it/paasify',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Cafram

Also known as: Pshiiit' Knackie ! Framework
"""
# pylint: disable=too-few-public-methods
# pylint: disable=unused-argument

import logging
import textwrap

from cafram.utils import serialize

_log = logging.getLogger(__name__)


# See: https://stackoverflow.com/questions/328851/printing-all-instances-of-a-class


# =====================================================================
# Exceptions
# =====================================================================


class CaframException(Exception):
    """Generic cafram exception"""


class DictExpected(CaframException):
    """Raised a dict was expected"""


class ListExpected(CaframException):
    """Raised a list was expected"""


class MissingIdent(CaframException):
    """Raised ident is not set"""


class NotExpectedType(CaframException):
    """Raised when types mismatchs"""


class ClassExpected(CaframException):
    """Raised when a class was expected"""


class InvalidSyntax(CaframException):
    """Raised when config string are not correct"""


class SchemaError(CaframException):
    """Raised when config does not match to schema"""


class ApplicationError(CaframException):
    """Raised when the developper introduced a bug"""


# =====================================================================
# Class helpers
# =====================================================================


class Base:
    """Base cafram object class

    Provides basic features such as:
    * module name
    * object kind
    * object name
    * object logger

    Available methods:
    * dump()
    """

    # Public attributes
    # ---------------------

    # Current library name
    module = "cafram"

    # Objects can have names
    ident = None

    # Object kind, nice name to replace raw class name, should be a string
    kind = None

    # Define live runtime data
    # To be renamed: shared
    # runtime = {}

    # Optional attributes
    # ---------------------

    # Object shortcut to logger
    log = None
    _node_log = None

    def __init__(self, *args, **kwargs):

        self.kind = kwargs.get("kind") or self.kind or self.__class__.__name__

        # Ident management
        if "ident" in kwargs:
            self.ident = kwargs.get("ident")
        if self.ident is None:
            raise MissingIdent(f"Missing 'ident' for __init__: {self}, {self.__dict__}")

        self.ident = self.ident or kwargs.get("ident") or self.kind

        self.log = kwargs.get("log") or self.log or logging.getLogger(__name__)
        self.shared = kwargs.get("shared") or {}

    def __str__(self):
        "Return a nice str representation of object"
        return f"{self.__class__.__name__}:{self.ident}"

    def __repr__(self):
        "Return a nice representation of object"
        # return self._nodes
        return f"{self.__class__.__name__}.{id(self)} {self.ident}"
        # return f"Instance {id(self)}: {self.__class__.__name__}:{self.ident}"

    # pylint: disable=redefined-builtin
    def dump(self, format="json", filter=None, **kwargs):
        "Show a dump of an object"

        print("\n===================================================================")
        print(f"== Dump of {self.module}.{self.kind}.{self.ident} {id(self)}")
        print("===================================================================\n")

        print("  Infos:")
        print("  -----------------")
        print(f"    ID: {id(self)}")
        print(f"    Kind: {self.kind}")
        print(f"    Ident: {self.ident}")
        print(f"    Repr: {repr(self)}")
        print(f"    String: {str(self)}")
        classes = "-> ".join([x.__name__ for x in self.__class__.__mro__])
        print(f"    MRO: {classes}")

        print("")


# =====================================================================
# Pre Base Class helpers
# =====================================================================


# Logging
# --------------------------


class MixInLog:
    "Provide a per instance logging"

    log = _log
    module = "cafram"
    conf_logger = None

    def __init__(self, *args, **kwargs):

        self.conf_logger = kwargs.get("log") or self.conf_logger
        self.set_logger()

    def set_logger(self, conf_logger=None, attribute_name='log'):
        """Set instance logger name or instance

        If never called, the default logger the cafram one

        attribute_name:
        * The attribute where to store the logger
        conf_logger:
        * None: Use default logging.getLogger()
        * "__AUTO__": Set to {self.module}.{self.__class__.__name__}.{ident}, and skip ident if absent
        * str: The name of your logger, as string
        * Logger: Directly a logger instance
        """

        log = None
        log_name = None
        conf_logger = conf_logger or self.conf_logger

        if conf_logger is None:
            log_name = f"{self.module}.{self.__class__.__name__}"
            ident = getattr(self, "ident", None)
            if ident:
                log_name = f"{self.module}.{self.__class__.__name__}.{ident}"
        elif isinstance(conf_logger, str):
            log_name = f"{conf_logger}"
        elif log.__class__.__name__ == "Logger":
            log = conf_logger

        else:
            raise Exception(f"Log not allowed here: {conf_logger}")

        if not log:
            log = logging.getLogger(log_name)

        # print ("Create logger", log_name, conf_logger, self.conf_logger)
        setattr(self, attribute_name, log)


# =====================================================================
# Post Base Class helpers
# =====================================================================


# Hooks
# --------------------------


class Hooks:
    """Provides a _init hook feature

    DEPRECATED
    """

    def __init__(self, *args, **kwargs):

        # pylint: disable=super-with-arguments
        # super(Hooks, self).__init__(*args, **kwargs)

        if callable(getattr(self, "_init", None)):
            self._init(*args, **kwargs)


# Family
# --------------------------


class Family(Base):
    "Provides a family tree"

    root = None
    parent = None
    children = []

    def __init__(self, *args, **kwargs):

        # pylint: disable=super-with-arguments
        super(Family, self).__init__(*args, **kwargs)

        # Init family
        parent = kwargs.get("parent") or self.parent

        # Register parent
        if parent and parent != self:
            self.parent = parent
            self.root = parent.root
        else:
            self.parent = self
            self.root = self

        # Register children
        self.children = []
        if self.has_parents():
            self.parent.children.append(self)

    def get_children_tree(self):
        "Get children tree"

        result = []
        children = self.children or []
        for child in children:
            children = child.get_children_tree()
            result.append({str(child): children or None})

        return result

    def has_parents(self):
        "Check if has parents"
        return bool(self.parent and self.parent != self)

    def get_parent(self):
        "Return parent"
        return self.parent or None

    def get_parents(self):
        "Return all parent of the object"

        parents = []
        current = self
        parent = self.parent or None
        while parent is not None and parent != current:
            if parent not in parents:
                parents.append(parent)
                current = parent
                parent = getattr(current, "parent")

        return parents

    # pylint: disable=arguments-differ
    def dump(self, **kwargs):
        "Return a dump of the family tree"

        # pylint: disable=super-with-arguments
        super(Family, self).dump(**kwargs)

        parents = self.get_parents()
        children = self.get_children_tree()

        if parents or children:
            print("  Family:")
            print("  -----------------")
            print(f"    Parents/Children: {len(parents)}/{len(children)}")

        if parents:
            print("    Parents:")
            # print ("    -----------------")
            parents.reverse()
            parents = serialize(parents, fmt="yaml")
            print(textwrap.indent(parents, "      "))

        if children:
            print("    Children:")
            # print ("    -----------------")
            children = serialize(children, fmt="yaml")
            print(textwrap.indent(children, "      "))

        print("\n")



# -*- coding: utf-8 -*-
"Paasify Framework Libary"


# pylint: disable=logging-fstring-interpolation

import os
import logging


from pprint import pprint  # noqa: F401


from cafram.nodes import NodeList, NodeMap, NodeDict
from cafram.base import MixInLog, Base

from cafram.utils import (
    # to_domain,
    first,
    merge_dicts,
)

import paasify.errors as error
from paasify.common import filter_existing_files

_log = logging.getLogger()
_first = first


class PaasifyObj(Base, MixInLog):
    "Default Paasify base object"

    module = "paasify.cli"
    conf_logger = None
    log = _log

    def __init__(self, *args, **kwargs):

        # Manually load classe
        Base.__init__(self, *args, **kwargs)
        MixInLog.__init__(self, *args, **kwargs)

        # Other things
        # super().__init__(*args, **kwargs)


class PaasifySimpleDict(NodeMap, PaasifyObj):
    "Simple Paaisfy Configuration Dict"

    conf_default = {}


class PaasifyConfigVar(NodeMap, PaasifyObj):
    "Simple Paaisfy Configuration Var Dict"

    conf_ident = "{self.name}={self.value}"
    conf_default = {
        "name": None,
        "value": None,
    }

    # conf_logger = "paasify.cli.ConfigVar"

    conf_schema = {
        "title": "Variable definition as key/value",
        "description": "Simple key value variable declaration, under the form of: {KEY: VALUE}. This does preserve value type.",
        "type": "object",
        "propertyNames": {"pattern": "^[A-Za-z_][A-Za-z0-9_]*$"},
        "minProperties": 1,
        "maxProperties": 1,
        # "additionalProperties": False,
        "patternProperties": {
            "^[A-Za-z_][A-Za-z0-9_]*$": {
                # ".*": {
                "title": "Environment Key value",
                "description": "Value must be serializable type",
                # "oneOf": [
                #    {"title": "As string", "type": "string"},
                #    {"title": "As boolean", "type": "boolean"},
                #    {"title": "As integer", "type": "integer"},
                #    {
                #        "title": "As null",
                #        "description": "If set to null, this will remove variable",
                #        "type": "null",
                #    },
                # ],
            }
        },
    }

    def node_hook_transform(self, payload):

        result = None
        if isinstance(payload, str):
            value = payload.split("=", 2)
            result = {
                "name": value[0],
                "value": value[1],
            }
        if isinstance(payload, dict):
            if "name" in payload and "value" in payload and len(payload.keys()) == 2:
                result = {
                    "name": payload["name"],
                    "value": payload["value"],
                }

            elif len(payload.keys()) == 1:
                for key, value in payload.items():
                    result = {
                        "name": key,
                        "value": value,
                    }

        if result is None:
            raise Exception(f"Unsupported type {type(payload)}: {payload}")

        return result


vardef_schema_complex = {
    "description": "Environment configuration. Paasify leave two choices for the configuration, either use the native dict configuration or use the docker-compatible format",
    "oneOf": [
        merge_dicts(
            PaasifyConfigVar.conf_schema,
            {
                "examples": [
                    {
                        "env": [
                            {"MYSQL_ADMIN_USER": "MyUser"},
                            {"MYSQL_ADMIN_DB": "MyDB"},
                        ]
                    },
                ],
            },
        ),
        {
            "title": "Variable definition as string (Compat)",
            "description": "Value must be a string, under the form of: KEY=VALUE. This does not preserve value type.",
            "type": "string",
            "pattern": "^[A-Za-z_][A-Za-z0-9_]*=.*$",
            "examples": [
                {
                    "env": [
                        "MYSQL_ADMIN_USER=MyUser",
                        "MYSQL_ADMIN_DB=MyDB",
                    ]
                },
            ],
        },
    ],
}


class PaasifyConfigVars(NodeList, PaasifyObj):
    "Paasify Project configuration object"

    conf_children = PaasifyConfigVar

    conf_schema = {
        # "$schema": "http://json-schema.org/draft-07/schema#",
        "title": "Environment configuration",
        "description": (
            "Environment configuration. Paasify leave two choices for the "
            "configuration, either use the native dict configuration or use the "
            "docker-compatible format"
        ),
        "oneOf": [
            {
                "title": "Env configuration as list",
                "description": (
                    "Configure variables as a list. This is the recommended way as"
                    "it preserves the variable parsing order, useful for templating. This format "
                    "allow multiple configuration format."
                ),
                "type": "array",
                "default": [],
                "additionalProperties": vardef_schema_complex,
                "examples": [
                    {
                        "env": [
                            # "MYSQL_ADMIN_DB=MyDB",
                            {"MYSQL_ADMIN_USER": "MyUser"},
                            {"MYSQL_ADMIN_DB": "MyDB"},
                            {"MYSQL_ENABLE_BACKUP": True},
                            {"MYSQL_BACKUPS_NODES": 3},
                            {"MYSQL_NODE_REPLICA": None},
                            "MYSQL_WELCOME_STRING=Is alway a string",
                        ],
                    },
                ],
            },
            {
                "title": "Env configuration as dict (Compat)",
                "description": (
                    "Configure variables as a dict. This option is only proposed for "
                    "compatibility reasons. It does not preserve the order of the variables."
                ),
                "type": "object",
                "default": {},
                # "patternProperties": {
                #    #".*": { "properties": PaasifyConfigVar.conf_schema, }
                #    ".*": { "properties": vardef_schema_complex , }
                # },
                "propertyNames": {"pattern": "^[A-Za-z_][A-Za-z0-9_]*$"},
                # "additionalProperties": PaasifyConfigVar.conf_schema,
                "examples": [
                    {
                        "env": {
                            "MYSQL_ADMIN_USER": "MyUser",
                            "MYSQL_ADMIN_DB": "MyDB",
                            "MYSQL_ENABLE_BACKUP": True,
                            "MYSQL_BACKUPS_NODES": 3,
                            "MYSQL_NODE_REPLICA": None,
                        }
                    },
                ],
            },
            {
                "title": "Unset",
                "description": ("Do not define any vars"),
                "type": "null",
                "default": None,
                "examples": [
                    {
                        "env": None,
                    },
                    {
                        "env": [],
                    },
                    {
                        "env": {},
                    },
                ],
            },
        ],
    }

    def node_hook_transform(self, payload):

        result = []
        if not payload:
            pass
        elif isinstance(payload, dict):
            for key, value in payload.items():
                var_def = {key: value}
                result.append(var_def)
        elif isinstance(payload, list):
            result = payload

        # elif isinstance(payload, str):
        #         value = payload.split("=", 2)
        #         result = {
        #             "name": value[0],
        #             "value": value[1],
        #         }
        else:
            raise error.InvalidConfig(f"Unsupported type: {payload}")

        return result

    def get_vars(self, current=None):
        "Parse vars and interpolate strings"

        result = dict(current or {})

        for var in self._nodes:
            value = var.value
            result[var.name] = value
        return result

    def get_vars_list(self, current=None):
        "Parse vars and interpolate strings"

        assert isinstance(current, list) or current is None

        result = list(current or [])
        for var in self._nodes:
            result.append(var)
        return result


class PaasifySource(NodeDict, PaasifyObj):
    "Paasify source configuration"

    def install(self, update=False):
        "Install a source if not updated"

        # Check if the source if installed or install latest

        prj = self.get_parent().get_parent()
        coll_dir = prj.runtime.project_collection_dir
        src_dir = os.path.join(coll_dir, self.ident)
        git_dir = os.path.join(src_dir, ".git")

        if os.path.isdir(git_dir) and not update:
            self.log.debug(
                f"Collection '{self.ident}' is already installed in: {git_dir}"
            )
            return

        self.log.info(f"Install source '{self.ident}' in: {git_dir}")
        raise NotImplementedError


class PaasifySources(NodeDict, PaasifyObj):
    "Sources manager"
    conf_children = PaasifySource


class FileReference:
    """A FileReference Object

    Useful for managing project paths

    The path, once created is immutable, you choose how it behave one time
    and done forever. They act a immutable local variable.


    path: The path you want to manage
    root: The root of your project, CWD else
    keep: Returned path will be returned as absolute
        True: Return default path from origin abs/rel
        False: Return default path from root_path abs/rel

    """

    def __init__(self, path, root=None, keep=False):

        assert isinstance(path, str), f"Got: {type(path)}"
        root = root or os.getcwd()
        self.raw = path
        self.root = root
        self.keep = keep

    def __str__(self):
        return self.path()

    def is_abs(self):
        "Return true if the path is absolute"
        return os.path.isabs(self.raw)

    def is_root_abs(self):
        "Return true if the root path is absolute"
        return os.path.isabs(self.root)

    def path(self, start=None):
        "Return the absolute or relative path from root depending if root is absolute or not"

        if self.keep:
            if self.is_abs():
                return self.path_abs(start=start)
            return self.path_rel(start=start)
        else:
            if self.is_root_abs():
                return self.path_abs(start=start)
            return self.path_rel(start=start)

    def path_abs(self, start=None):
        "Return the absolute path from root"

        if self.is_abs():
            result = self.raw
        else:
            start = start or self.root
            real_path = os.path.join(start, self.raw)
            result = os.path.abspath(real_path) or "."
        return result

    def path_rel(self, start=None):
        "Return the relative path from root"

        if self.is_abs():
            start = start or self.root
            result = os.path.relpath(self.raw, start=start)
        else:
            start = start or self.root
            real_path = os.path.join(start, self.raw)
            result = os.path.relpath(real_path) or "."
        return result


class FileLookup(PaasifyObj):
    """A FileLookup Object

    Useful for identifing available files in differents hierachies"""

    def __init__(self, path=None, pattern=None, **kwargs):
        self._lookups = []
        self.log = logging.getLogger("paasify.cli.FileLookup")

        if path and pattern:
            self.insert(path, pattern, **kwargs)

    def _parse(self, path, pattern, **kwargs):
        "Ensure lookup is correctly formed"
        data = dict(kwargs)
        data.update(
            {
                "path": path,
                "pattern": pattern if isinstance(pattern, list) else [pattern],
            }
        )
        return data

    def insert(self, path, pattern, **kwargs):
        "Insert first a path/pattern to the lookup object"
        data = self._parse(path, pattern, **kwargs)
        self._lookups.insert(0, data)

    def append(self, path, pattern, **kwargs):
        "Append a path/pattern to the lookup object"
        data = self._parse(path, pattern, **kwargs)
        self._lookups.append(data)

    def get_lookups(self):
        "Return object lookups"
        return self._lookups

    def lookup_candidates(self):
        "List all available candidates of files for given folders, low level"

        result = []
        for lookup in self._lookups:
            path = lookup["path"]
            if path:
                cand = filter_existing_files(path, lookup["pattern"])
                lookup["matches"] = cand
                result.append(lookup)
        return result

    def paths(self, first=False):
        "All matched files"

        vars_cand = self.lookup_candidates()
        ret = []
        for cand in vars_cand:
            for match in cand["matches"]:
                ret.append(match)

        if first:
            return _first(ret) if len(ret) > 0 else None
        return ret

    def match(self, fail_on_missing=False, first=False):
        "Match all candidates, and built a list result with object inside"

        vars_cand = self.lookup_candidates()
        result = []
        missing = []
        for cand in vars_cand:
            matches = cand["matches"]
            for match in matches:
                payload = dict(cand)
                payload.update({"match": match})
                result.append(payload)
            if len(matches) == 0:
                # Still add unlucky candidates
                missing.append(cand)

        # Report errors if missing
        if fail_on_missing and len(missing) > 0:
            missing_paths = [
                os.path.join(lookup["path"], first(lookup["pattern"]))
                for lookup in missing
            ]
            for missed in missing_paths:
                self.log.info(f"Missing file: {missed}")
            missed_str = ",".join(missing_paths)
            raise error.MissingFile(
                f"Can't load {len(missing_paths)} vars files: {missed_str}"
            )

        if first:
            return _first(result) if len(result) > 0 else None

        return result



# -*- coding: utf-8 -*-
"""
Project library

This handle the project entity.

This library provides two classes:

* PaasifyProjectConfig: A project config
* PaasifyProject: A project instance


Example:
``` py title="test.py"
from paasify.projects import PaasifyProject

prj = PaasifyProject.discover_project()
prj.dump()
```
"""

# pylint: disable=logging-fstring-interpolation


import os

from pprint import pprint  # noqa: F401
import anyconfig

from cafram.nodes import NodeMap, NodeList

import paasify.errors as error
from paasify.engines import EngineDetect
from paasify.sources import SourcesManager
from paasify.framework import (
    PaasifyObj,
    PaasifyConfigVars,
)
from paasify.common import list_parent_dirs, find_file_up, get_paasify_pkg_dir
from paasify.stacks import StackTagMgr, StackManager


ALLOW_CONF_JUNK = False


class PaasifyConfigExtraVars(NodeList, PaasifyObj):
    "Paasify Project Configuration"

    conf_default = []

    conf_schema = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "title": "Project extra vars files",
        "description": "Extra vars file to load for this project",
        "oneOf": [
            {
                "title": "List of file",
                "description": "List of files to load",
                "type": "array",
                "items": {
                    "type": "string",
                },
            },
            {
                "title": "String",
                "description": "A single file to load",
                "type": "string",
            },
            {
                "title": "No configuration",
                "description": "Disable extra_vars loading",
                "type": "null",
            },
        ],
        "examples": [
            {
                "extra_vars": [
                    "../common_vars.yml"
                    "common/common_vars.yml"
                    "/absolute/path/common_vars.yml"
                ],
            },
            {
                "extra_vars": "single/file/to_load.yml",
            },
            {
                "extra_vars": None,
            },
        ],
    }

    def node_hook_transform(self, payload):
        """Init PassifyRuntime"""

        if not payload:
            payload = []
        return payload


class PaasifyProjectConfig(NodeMap, PaasifyObj):
    "Paasify Project Configuration"

    conf_default = {
        "namespace": None,
        "vars": {},
        "tags": [],
        "tags_suffix": [],
        "tags_prefix": [],
        "extra_vars": [],
    }

    conf_children = [
        {
            "key": "namespace",
        },
        {
            "key": "vars",
            "cls": PaasifyConfigVars,
        },
        {
            "key": "extra_vars",
            "cls": PaasifyConfigExtraVars,
        },
        # {
        #     "key": "tags",
        #     "cls": list,
        #     #"cls": StackTagMgr,
        # },
        # {
        #     "key": "tags_prefix",
        #     "cls": list,
        #     #"cls": StackTagMgr,
        # },
        # {
        #     "key": "tags_suffix",
        #     "cls": list,
        #    # "cls": StackTagMgr,
        # },
    ]

    conf_schema = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "title": "Paasify Project settings",
        "description": (
            "Configure main project settings. It provides global settings"
            " but also defaults vars and tags for all stacks."
        ),
        "oneOf": [
            {
                "type": "object",
                "additionalProperties": ALLOW_CONF_JUNK,
                "title": "Project configuration",
                "description": (
                    "Configure project as a dict value. "
                    "Most of these settings are overridable via environment vars."
                ),
                "default": {},
                "properties": {
                    "namespace": {
                        "title": "Project namespace",
                        "description": (
                            "Name of the project namespace. If not"
                            " set, defaulted to directory name"
                        ),
                        "oneOf": [
                            {
                                "title": "None",
                                "description": "Defaulted by the project dir name",
                                "type": "null",
                            },
                            {
                                "title": "String",
                                "description": "Custom namespace name string",
                                "type": "string",
                            },
                        ],
                    },
                    "extra_vars": PaasifyConfigExtraVars.conf_schema,
                    "vars": PaasifyConfigVars.conf_schema,
                    "tags": StackTagMgr.conf_schema,
                    "tags_suffix": StackTagMgr.conf_schema,
                    "tags_prefix": StackTagMgr.conf_schema,
                },
                "examples": [
                    {
                        "config": {
                            "namespace": "my_ns1",
                            "vars": [{"my_var1": "my_value1"}],
                            "tags": ["tag1", "tag2"],
                        },
                    }
                ],
            },
            {
                "type": "null",
                "title": "Empty",
                "description": "Use automatic conf if not set. You can still override conf values with environment vars.",
                "default": None,
                "examples": [
                    {
                        "config": None,
                    },
                    {
                        "config": {},
                    },
                ],
            },
        ],
    }


class PaasifyProjectRuntime(NodeMap, PaasifyObj):
    "Paasify Runtime Object (deprecated)"

    conf_schema = {
        # TODO: "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "title": "Paasify Project settings",
        "additionalProperties": False,
        "properties": {
            "default_source": {
                "title": "",
                "description": "",
                "type": "string",
            },
            "cwd": {
                "title": "",
                "description": "",
                "type": "string",
            },
            "working_dir": {
                "title": "",
                "description": "",
                "oneOf": [
                    {"type": "string"},
                    {"type": "null"},
                ],
            },
            "engine": {
                "title": "Docker backend engine",
                "oneOf": [
                    {
                        "description": "Docker engine",
                        "type": "string",
                    },
                    {
                        "description": "Automatic",
                        "type": "null",
                    },
                ],
            },
            "filenames": {
                "oneOf": [
                    {
                        "title": "List of file to lookup",
                        "description": "List of string file names to lookup paasify.yaml files",
                        "type": "array",
                        "items": {
                            "type": "string",
                        },
                    },
                ],
            },
        },
    }

    conf_default = {
        "load_file": None,
        "root_hint": None,
        "default_source": "default",
        "cwd": os.getcwd(),
        "working_dir": None,
        "engine": None,
        "filenames": ["paasify.yml", "paasify.yaml"],
        "relative": None,
        "dump_payload_log": False,
        "no_tty": False,
    }

    def node_hook_transform(self, payload):
        """Init PassifyRuntime"""

        # Allow config as string !
        if isinstance(payload, str):
            root_hint = payload
            payload = {
                "root_hint": root_hint,
                "load_file": True,
            }

        # Create default config
        result = {}
        result = dict(self.conf_default)
        result.update(payload)

        # The payload is a dir or a config file
        root_hint = result.get("root_hint")
        filenames = result.get("filenames")
        _payload1 = self.get_ctx(root_hint, config_files=filenames)
        result.update(_payload1)

        # Build default runtime from root path
        root_path = result["root_path"]

        paasify_source_dir = get_paasify_pkg_dir()
        paasify_plugins_dir = os.path.join(paasify_source_dir, "assets", "plugins")
        private_dir = os.path.join(root_path, ".paasify")
        collection_dir = os.path.join(private_dir, "collections")
        jsonnet_dir = os.path.join(private_dir, "plugins")

        _payload2 = {
            "paasify_source_dir": paasify_source_dir,
            "paasify_plugins_dir": paasify_plugins_dir,
            "project_private_dir": private_dir,
            "project_collection_dir": collection_dir,
            "project_jsonnet_dir": jsonnet_dir,
        }
        result.update(_payload2)

        # Allow user to override parts
        result.update(payload)
        return result

    @classmethod
    def get_project_path2(cls, path, filenames=None):
        "Find the closest paasify config file"

        # if not path.startswith('/'):

        filenames = filenames or cls.filenames
        # filenames = self._node_root.config.filenames

        paths = list_parent_dirs(path)
        result = find_file_up(filenames, paths)

        return result

    @staticmethod
    def get_ctx(project_hint, config_files=None, cwd=None, relative=None):
        "Return a list of directory context"

        config_files = config_files or ["paasify.yml", "paasify.yaml"]
        cwd = cwd or os.getcwd()

        # Autofind config file in parents if None
        if project_hint is None:
            # Show relative by default when project_hint is None
            relative = True if relative is None else relative

            try:
                project_hint = PaasifyProjectRuntime.get_project_path2(
                    cwd, filenames=config_files
                )[0]
            except IndexError as err:
                config_files = "' or '".join(config_files)
                msg = f"Impossible to find any '{config_files}' in '{cwd}', or in above directories."
                raise error.ProjectNotFound(msg) from err

        # Check the project root:
        if os.path.isdir(project_hint):
            root_path = project_hint
            # TODO: Lookup for candidates instead taking the first
            config_file_name = config_files[0]

        elif os.path.isfile(project_hint):
            root_path = os.path.dirname(project_hint)
            config_file_name = os.path.basename(project_hint)
        else:
            msg = f"Impossible to find paasify project in: {project_hint}"
            raise error.ProjectNotFound(msg)

        assert root_path

        # Get more context
        if relative is None:
            relative = not os.path.isabs(project_hint)
        project_rel = os.path.relpath(root_path, start=cwd)
        project_abs = os.path.abspath(root_path)

        # Check if cwd inside
        sub_dir = None
        if project_abs != cwd and project_abs in cwd:
            sub_dir = cwd.replace(project_abs, "").strip("/")
        # elif project_abs != cwd:
        #     relative = False

        # Convert root_path
        root_path = project_rel if relative else project_abs

        result = {
            "namespace": os.path.basename(project_abs),
            "root_path": root_path,
            "config_file": config_file_name,
            "config_file_path": os.path.join(root_path, config_file_name),
            "relative": relative,
            "cwd": cwd,
            "sub_dir": sub_dir,
        }
        return result


class PaasifyProject(NodeMap, PaasifyObj):
    "Paasify Project instance"

    conf_default = {
        "_runtime": {},
        "config": {},
        "sources": [],
        "stacks": [],
    }

    conf_children = [
        {
            "key": "config",
            "cls": PaasifyProjectConfig,
        },
        {
            "key": "sources",
            "cls": SourcesManager,
        },
        {
            "key": "stacks",
            "cls": StackManager,
        },
    ]

    conf_schema = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "title": "Paasify",
        "description": "Main paasify project settings. This defines the format of `paasify.yml`.",
        "additionalProperties": ALLOW_CONF_JUNK,
        "examples": [
            {
                "config": {
                    "tags_prefix": ["_paasify"],
                    "vars": {
                        "app_domain": "devbox.192.168.186.129.nip.io",
                        "app_expose_ip": "192.168.186.129",
                        "app_tz": "Europe/Paris",
                        "top_var1": "My value",
                        "top_var2": "TOP VAR1=> ${top_var1}",
                    },
                },
                "sources": [
                    {"default": {"url": "https://github.com/user/docker-compose.git"}}
                ],
                "stacks": [
                    {
                        "app": "default:traefik",
                        "path": "traefik",
                        "tags": [
                            "ep_http",
                            "expose_admin",
                            "debug",
                            {
                                "traefik-svc": {
                                    "traefik_net_external": False,
                                    "traefik_svc_port": "8080",
                                }
                            },
                        ],
                    },
                    {
                        "app": "default:minio",
                        "env": [
                            {"app_admin_passwd": "MY_PASS"},
                            {"app_image": "quay.io/minio/minio:latest"},
                        ],
                        "tags": [
                            {
                                "traefik-svc": {
                                    "traefik_svc_name": "minio-api",
                                    "traefik_svc_port": 9000,
                                }
                            },
                            {
                                "traefik-svc": {
                                    "traefik_svc_name": "minio-console",
                                    "traefik_svc_port": 9001,
                                }
                            },
                        ],
                    },
                    {"app": "default:authelia", "tags": ["traefik-svc"]},
                    {"app": "default:librespeed", "tags": ["traefik-svc"]},
                ],
            },
        ],
        "default": {},
        "properties": {
            "config": {
                "type": "object",
                "description": "See: schema prj_config",
            },
            "sources": {
                "type": "object",
                "description": "See: schema prj_sources",
            },
            "stacks": {
                "type": "array",
                "description": "See: schema prj_stacks",
            },
            "_runtime": {
                "title": "Project runtime variables",
                "type": "object",
                "description": "Internal object to pass context variables, internal use only or for troubleshooting purpose",
            },
        },
    }

    ident = "main"
    engine_cls = None
    runtime = None

    def node_hook_transform(self, payload):
        "Init configuration Project"

        # Create runtime instance child
        _runtime = payload.get("_runtime") or payload
        self.runtime = PaasifyProjectRuntime(
            parent=self, payload=_runtime, ident="ProjectRuntime"
        )

        # Inject payload
        if self.runtime.load_file is not False:
            self.log.info(f"Load project file: {self.runtime.config_file_path}")
            _payload = anyconfig.load(self.runtime.config_file_path)
            payload.update(_payload)

        # Create engine
        if not self.engine_cls:
            engine_name = self.runtime.engine or None
            self.engine_cls = EngineDetect().detect(engine=engine_name)

        return payload

    def node_hook_final(self):
        "Report for logging"
        ns = self.runtime.namespace
        self.log.info(f"Project '{ns}' loaded")
